// g++ -D RESIZINGARRAYDEQUE_MAIN -std=c++0x ResizingArrayDeque.cpp
// g++ -D RESIZINGARRAYDEQUE_MAIN -std=c++11 ResizingArrayDeque.cpp


#include "ResizingArrayDeque.hpp"

#include <cassert>
#include <cstddef>

#include <iostream>

template <typename T>
ResizingArrayDeque<T>::ResizingArrayDeque( void ) :
    N (0),
    max(1),
    first(nullptr),
    last(nullptr) {
    data = new T[2];
    }
    
template <typename T>
ResizingArrayDeque<T>::~ResizingArrayDeque( void ) {
    assert( N == 0 );  // Catch the potential memory leak.
    if ( data )
        delete[] data;
    }

template <typename T>
void ResizingArrayDeque<T>::push_right( T &item ) {
    if ( N == max )
        resize( 2 * max );

    if ( last ) {
        if ( last == data )
            last = data + max - 1;
        else
            ++last;
        }
    else {
        first = last = data;
        }

    *last = item;
    ++N;
    }

template <typename T>
void ResizingArrayDeque<T>::push_left( T &item ) {
    if ( N == max )
        resize( 2 * max );

    if ( first ) {
        if ( first == data )
            first = data + max - 1;
        else
            --first;
        }
    else {
        first = last = data;
        }

    *first = item;
    ++N;
    }

template <typename T>
T ResizingArrayDeque<T>::pop_right( void ) {
    assert( N > 0 );

    T item = *last;
    N--;

    if ( last == data ) {
        last = data + max - 1;
        }
    else {
        --last;
        }

    if ( N == 0 )
        first = last = nullptr;

    // if ( N <= max/4 )
    //     resize( max/2 );

    return item;
    }

template <typename T>
T ResizingArrayDeque<T>::pop_left( void ) {
    assert( N > 0 );

    T item = *first;
    N--;

    if ( first == data + max - 1 ) {
        first = data;
        }
    else {
        ++first;
        }

    if ( N == 0 )
        first = last = nullptr;

    // if ( N <= max/4 )
    //     resize( max/2 );

    return item;
    }

template <typename T>
bool ResizingArrayDeque<T>::is_empty( void ) {
    return N == 0;
    }

template <typename T>
size_t ResizingArrayDeque<T>::size( void ) {
    return N;
    }

template <typename T>
void ResizingArrayDeque<T>::resize( size_t new_max ) {
    T *new_data = new T[new_max];

    if ( N ) {
        size_t i = 0;
        T *curr = first;
        while ( i < N ) {
            new_data[i] = *curr;
            ++curr;
            if ( curr == data + max )
                curr = data;
            ++i;
            }
        }
    first = new_data;
    last = new_data + N - 1;

    T *old_data = data;
    data = new_data;
    delete[] old_data;

    max = new_max;
    }

template <typename T>
typename ResizingArrayDeque<T>::iterator ResizingArrayDeque<T>::begin( void ) {
    return ResizingArrayDeque<T>::iterator( first, last, data, data + max - 1 );
    }

template <typename T>
typename ResizingArrayDeque<T>::iterator ResizingArrayDeque<T>::end( void ) {
    return ResizingArrayDeque<T>::iterator( nullptr, nullptr, nullptr, nullptr );
    }


////////////////////////////////////////////////////////////////////////////////

template <typename T>
ResizingArrayDeque<T>::iterator::iterator( T *c, T *l, T *b, T *m ) :
    curr (c),
    last (l),
    begin (b),
    max (m)
    {
    }

template <typename T>
typename ResizingArrayDeque<T>::iterator & ResizingArrayDeque<T>::iterator::operator++() {
    if ( this->curr == this->last ) {
        this->curr = nullptr;
        }
    else {
        this->curr += 1;
        if ( this->curr > this->max )
            this->curr = this->begin;
        }

    return *this;
    }

template <typename T>
typename ResizingArrayDeque<T>::iterator ResizingArrayDeque<T>::iterator::operator++( int ) {
    auto t = ResizingArrayDeque<T>::iterator( *this );

    if ( this->curr == this->last ) {
        this->curr = nullptr;
        }
    else {
        this->curr += 1;
        if ( this->curr > this->max )
            this->curr = this->begin;
        }

    return t;
    }

template <typename T>
T ResizingArrayDeque<T>::iterator::operator*() {
    return *(this->curr);
    }

template <typename T>
bool ResizingArrayDeque<T>::iterator::operator!=( typename ResizingArrayDeque<T>::iterator other ) {
    return this->curr != other.curr;
    }


////////////////////////////////////////////////////////////////////////////////

#ifdef RESIZINGARRAYDEQUE_MAIN

#include <iostream>

int main ( int argc, char **argv ) {

    ResizingArrayDeque<long> deque;
    bool left = true;

    long i;
    while ( ! std::cin.eof() ) {
        std::cin >> i;
        if ( std::cin.good() ) {
            if ( left ) {
                deque.push_left(i);
                left = false;
                }
            else {
                deque.push_right(i);
                left = true;
                }
            }
        }

    std::cout << "Deque has " << deque.size() << " entries." << std::endl;

    for ( auto iter = deque.begin(); iter != deque.end(); ++iter ) {
        std::cout << *iter << std::endl;
        }

    std::cout << "Popping entries..." << std::endl;

    left = true;
    while ( ! deque.is_empty() ) {
        if ( left ) {
            i = deque.pop_left();
            left = false;
            }
        else {
            i = deque.pop_right();
            left = true;
            }
        std::cout << i << std::endl;
        }

    std::cout << "Deque has " << deque.size() << " entries." << std::endl;

    }

#endif
