// g++ -D STEQUE_MAIN -std=c++0x Steque.cpp
// g++ -D STEQUE_MAIN -std=c++11 Steque.cpp


#include "Steque.hpp"

#include <cassert>
#include <cstddef>

#include <iostream>

template <typename T>
Steque<T>::Steque( void ) :
    N (0),
    list (nullptr)
    {
    }
    
template <typename T>
Steque<T>::~Steque( void ) {
    assert( N == 0 );  // Catch the potential memory leak.
    while ( list ) {
        Node * next = list->next;
        delete list;
        list = next;
        }
    }

template <typename T>
void Steque<T>::push( T &item ) {
    Node * node = new Node();
    node->item = item;
    node->next = list;
    list = node;
    N++;
    }

template <typename T>
T Steque<T>::pop( void ) {
    assert( N > 0 );

    T item = list->item;

    Node *old = list;
    list = list->next;
    delete old;
    N--;

    return item;
    }

template <typename T>
void Steque<T>::enqueue( T &item ) {
    Node *curr = list;

    Node *node = new Node();
    node->item = item;
    node->next = nullptr;

    if ( ! curr ) {
        list = node;
        N++;
        return;
        }

    while ( curr->next ) {
        curr = curr->next;
        }

    curr->next = node;
    N++;

    return;
    }

template <typename T>
bool Steque<T>::is_empty( void ) {
    return N == 0;
    }

template <typename T>
size_t Steque<T>::size( void ) {
    return N;
    }

template <typename T>
typename Steque<T>::iterator Steque<T>::begin( void ) {
    return Steque<T>::iterator( list );
    }

template <typename T>
typename Steque<T>::iterator Steque<T>::end( void ) {
    return Steque<T>::iterator( nullptr );
    }


////////////////////////////////////////////////////////////////////////////////

template <typename T>
Steque<T>::iterator::iterator( Node *c ) :
    curr (c)
    {
    }

template <typename T>
typename Steque<T>::iterator & Steque<T>::iterator::operator++() {
    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return *this;
    }

template <typename T>
typename Steque<T>::iterator Steque<T>::iterator::operator++( int ) {
    auto t = Steque<T>::iterator( *this );

    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return t;
    }

template <typename T>
T Steque<T>::iterator::operator*() {
    return this->curr->item;
    }

template <typename T>
bool Steque<T>::iterator::operator!=( typename Steque<T>::iterator other ) {
    return this->curr != other.curr;
    }

////////////////////////////////////////////////////////////////////////////////



#ifdef STEQUE_MAIN

#include <iostream>

int main ( int argc, char **argv ) {

    Steque<long> steque;
    bool push_it = true;

    long i;
    while ( ! std::cin.eof() ) {
        std::cin >> i;
        if ( std::cin.good() )  {
            if ( push_it ) {
                steque.push( i );
                push_it = false;
                }
            else {
                steque.enqueue( i );
                push_it = true;
                }
            }
        }

    std::cout << "Steque has " << steque.size() << " entries." << std::endl;

    for ( auto iter = steque.begin(); iter != steque.end(); ++iter ) {
        std::cout << *iter << std::endl;
        }

    std::cout << "Popping entries..." << std::endl;

    while ( ! steque.is_empty() ) {
        i = steque.pop();
        std::cout << i << std::endl;
        }

    std::cout << "Steque has " << steque.size() << " entries." << std::endl;

    }

#endif
